/*
 * Copyright (c) 2009 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 *
 * Author(s): Torsten Hoefler <htor@cs.indiana.edu>
 *
 */

#include "mpiparmetis.hpp"
#include <assert.h>
#include <vector>
#include <map>

typedef struct { int psrc,tgt; } twoint;


/* finds element i in array of ranges arr */
int find_part(int *arr, int size, int i) {
  /* binary search */
  int pos, min=0, max=size;

  do {
    pos=(min+max) / 2;
    if(i > arr[pos]) min = pos+1;
    else max = pos-1;
  } while((i != arr[pos]) && (min < max));

  /* the binary search doesn't necessarily find the right interval,
     however, it brings us close - should go more elegantly! */
  while( !(arr[pos] <= i && arr[pos+1] > i) ){
    if(arr[pos] > i) pos--;
    else pos++;
  }
  return pos;
}


int MPIX_Graph_create_parmetis_unweighted(idxtype *vtxdist, idxtype *xadj, idxtype *adjncy, 
    int *numflag, idxtype *part, MPI_Comm *comm, MPI_Comm *newcomm) {

  int r, p, ret /* ret is overwritten, be careful, the function only returns the status of the last call */;
  MPI_Comm_rank(*comm, &r);
  MPI_Comm_size(*comm, &p);

  /* correct indices -- each is '1' too big */
  if(*numflag == 1) {
    for(int i=0; i<p; i++) vtxdist[i] -= 1;
    for(int src=vtxdist[r]; src<vtxdist[r+1]; src++) {
      int srcind = src-vtxdist[r];
      xadj[srcind] -= 1;
      part[srcind] -= 1;
      srcind -= 1;
      for(int tgtind=xadj[srcind]; tgtind<xadj[srcind+1]; tgtind++) {
        adjncy[tgtind] -= 1;
      }
    }
  }

  int printrank=-1; // set negative to disable

  // counts for alltoallv
  std::vector<int> scounts(p,0), rcounts(p,0);

  // first pass over graph -- get all the counts for Alltoallv
  for(int src=vtxdist[r]; src<vtxdist[r+1]; src++) {
    int srcind = src-vtxdist[r];
    for(int tgtind=xadj[srcind]; tgtind<xadj[srcind+1]; tgtind++) {
      int tgt = adjncy[tgtind];
      int owner = find_part(vtxdist, p, tgt);
      //if(r==printrank) printf("[%i] %i -> %i (owner %i (%i-%i))\n", r, src, tgt, owner, vtxdist[owner], vtxdist[owner+1]);
      if(r != owner) { // target is remote
        scounts[owner]++; // send partition to owner
        rcounts[owner]++; // graph is symmetric (ParMETIS requires this), thus, receive mapping for target from owner of target
  } } }

  // fill displs arrays
  std::vector<int> sdispls(p), rdispls(p);
  sdispls[0]=rdispls[0] = 0;
  for(int i=0; i<p; i++) {
    sdispls[i]=sdispls[i-1]+scounts[i-1];
    rdispls[i]=rdispls[i-1]+rcounts[i-1];
  } 

  // buffers
  int ssize=sdispls[p-1]+scounts[p-1];
  int rsize=rdispls[p-1]+rcounts[p-1];
  std::vector<twoint> sbuf(ssize), rbuf(rsize);
  std::vector<int> sidx(p,0);

  // second pass over array -- initialize all buffers
  for(int src=vtxdist[r]; src<vtxdist[r+1]; src++) {
    int srcind = src-vtxdist[r];
    for(int tgtind=xadj[srcind]; tgtind<xadj[srcind+1]; tgtind++) {
      int tgt = adjncy[tgtind];
      int owner = find_part(vtxdist, p, tgt);
      if(r != owner) { // target is remote
        // push correctly mapped source vertex to owner
        sbuf[sdispls[owner]+sidx[owner]].psrc = part[srcind];
        sbuf[sdispls[owner]+sidx[owner]].tgt = tgt;
        sidx[owner]++;
  } } }

  // exchange remote edges
  //printf("[%i] %i %i %i %i - %i %i %i %i \n", r, scounts[0], scounts[1],scounts[2],scounts[3],rcounts[0],rcounts[1],rcounts[2],rcounts[3]);
  ret = MPI_Alltoallv(&sbuf[0], &scounts[0], &sdispls[0], MPI_2INT, &rbuf[0], &rcounts[0], &rdispls[0], MPI_2INT, *comm);

  std::vector<int> edgecnt(p,0);  // edge-count for the graph topology
  std::map<int,int> vert2idx;

  // third pass over array -- count numbers of vertices
  for(int src=vtxdist[r]; src<vtxdist[r+1]; src++) {
    int srcind = src-vtxdist[r];
    for(int tgtind=xadj[srcind]; tgtind<xadj[srcind+1]; tgtind++) {
      int tgt = adjncy[tgtind];
      int owner = find_part(vtxdist, p, tgt);
      if(r == owner) { // target is local
        int loc_idx=tgt-vtxdist[r];
        //printf("[%i] investigating local target %i->%i (%i->%i)\n", r, src, tgt, part[srcind], part[loc_idx]);
        int srcnode=part[srcind];
        // remote edge
        if(srcnode != part[loc_idx]) edgecnt[srcnode]++;
      }
    } 
    /* build index map of local vertices -- for faster access later --
     * btw., the third pass over the array could be overlapped with the
     * alltoall above ;-) */
    vert2idx[src] = srcind;
  }

  // loop over received edges array -- use the local vertex index map!
  for(int src=0; src<p; src++) {
    for(int i=rdispls[src]; i<rdispls[src]+rcounts[src]; ++i) {
      int srcnode=rbuf[i].psrc;
      int tgtnode=part[ vert2idx[ rbuf[i].tgt ] ];
      // remote edge
      if(srcnode != tgtnode) edgecnt[srcnode]++;
    }
  }

  // create array for graph topology
  int ngr=0; // number of remote vertices
  // get ngr
  for(int i=0; i<p; i++) if(edgecnt[i] > 0) ngr++;
  int edges=0; // number of outgoing edges
  int *degrees = (int*)malloc(ngr*sizeof(int)); // edgecnt for MPI interface
  int *edgeoff = (int*)malloc(p*sizeof(int)); // offsets of degrees (like CSR)
  int *sources = (int*)malloc(ngr*sizeof(int));
  int igr=0;
  for(int i=0; i<p; i++) {
    if(edgecnt[i] > 0) {
      degrees[igr]=edgecnt[i];
      sources[igr]=i;
      igr++; // next available vertex
    }
    edges += edgecnt[i];
    if(i==0) edgeoff[0]=0;
    else edgeoff[i]=edgeoff[i-1]+edgecnt[i-1];
  }
  if(r==printrank) { printf("[%i] edgeoff: ", r); for(int i=0; i<p; i++) printf("%i ", edgeoff[i]); printf("\n"); }

  int *destinations = (int*)malloc(edges*sizeof(int));

  // reset edgecnt
  for(int i=0; i<p; i++) edgecnt[i]=0; 

  // fouth and last pass over array -- create dist topology arrays
  for(int src=vtxdist[r]; src<vtxdist[r+1]; src++) {
    int srcind = src-vtxdist[r];
    for(int tgtind=xadj[srcind]; tgtind<xadj[srcind+1]; tgtind++) {
      int tgt = adjncy[tgtind];
      int owner = find_part(vtxdist, p, tgt);
      if(r == owner) { // target is local
        int loc_idx=tgt-vtxdist[r];
        //printf("[%i] investigating local target %i->%i (%i->%i)\n", r, src, tgt, part[srcind], part[loc_idx]);
        int srcnode=part[srcind];
        // remote edge
        if(srcnode != part[loc_idx]) {
          destinations[edgeoff[srcnode]+edgecnt[srcnode]] = part[loc_idx];
          edgecnt[srcnode]++;
        }
      }
    } 
  }

  // loop over received edges array -- use the local vertex index map!
  for(int src=0; src<p; src++) {
    for(int i=rdispls[src]; i<rdispls[src]+rcounts[src]; ++i) {
      int srcnode=rbuf[i].psrc;
      int tgtnode=part[ vert2idx[ rbuf[i].tgt ] ];
      // remote edge
      if(srcnode != tgtnode) {
        destinations[edgeoff[srcnode]+edgecnt[srcnode]] = tgtnode;
        edgecnt[srcnode]++;
      }
    }
  }

  if(r==printrank) {
    printf("sources: "); for(int i=0;i<ngr;i++) printf("%i ", sources[i]); printf("\n");
    printf("degrees: "); for(int i=0;i<ngr;i++) printf("%i ", degrees[i]); printf("\n");
    printf("destinations: "); { int cnt=0; for(int i=0;i<ngr;i++) for(int j=0; j<degrees[i];j++) { printf("%i ", destinations[cnt]); cnt++; } } printf("\n");
  }

//#define CHECK
#ifdef CHECK
  if(!r) printf("checking edges ...\n");
  {for(int i=0; i<ngr; ++i) if(sources[i] > p || sources[i] < 0) printf("[%i] xxx found illegal edge %i\n", r, sources[i]);}
  {int cnt=0; for(int i=0; i<ngr; ++i) for(int j=0; j<degrees[i];j++) { if(destinations[cnt] > p || destinations[cnt] < 0) printf("[%i] xxx found illegal edge %i\n", r, sources[i]); cnt++; } }
#endif

  ret = MPI_Dist_graph_create(*comm, ngr, sources, degrees, destinations, MPI_UNWEIGHTED, MPI_INFO_NULL, 1, newcomm);

  /* correct indices back -- each is '1' too small now, need to do it in
   * reverse order than at the beginning */
  if(*numflag == 1) {
    for(int i=0; i<p; i++) vtxdist[i] -= 1;
    for(int src=vtxdist[r]; src<vtxdist[r+1]; src++) {
      int srcind = src-vtxdist[r];
      for(int tgtind=xadj[srcind]; tgtind<xadj[srcind+1]; tgtind++) {
        adjncy[tgtind] += 1;
      }
    }
    for(int src=vtxdist[r]; src<vtxdist[r+1]; src++) {
      int srcind = src-vtxdist[r];
      xadj[srcind] += 1;
      part[srcind] += 1;
    }
    for(int i=0; i<p; i++) vtxdist[i] += 1;
  }

  return ret;
}
